package com.yanzuoguang.util.thread;

import com.yanzuoguang.util.log.Log;

import java.util.*;

/**
 * 执行队列的数据,会根据其中的执行函数，自动优化执行循序
 */
public final class RunnableListAuto {

    /**
     * 是否打印线程日志
     */
    public static boolean IsLog = false;

    /**
     * 缓存的执行对象的数据大小
     */
    private static Map<String, RunnableListAutoItem> cache = new Hashtable<String, RunnableListAutoItem>();

    /**
     * 自动开启线程去执行代码，并根据执行时间调整先后顺序
     *
     * @param codes
     */
    public static ThreadList<RunnableListAutoItem> Execute(Runnable... codes) {
        return Execute(Arrays.asList(codes));
    }

    /**
     * 自动开启线程去执行代码，并根据执行时间调整先后顺序
     *
     * @param codes
     */
    public static ThreadList<RunnableListAutoItem> Execute(List<Runnable> codes) {
        List<RunnableListAutoItem> autos = new ArrayList<RunnableListAutoItem>();
        for (Runnable code : codes) {
            autos.add(new RunnableListAutoItem(code));
        }
        return ExecuteAuto(autos);
    }

    /**
     * 自动开启线程去执行代码，并根据执行时间调整先后顺序
     *
     * @param methods
     */
    public static ThreadList<RunnableListAutoItem> ExecuteAuto(RunnableListAutoItem... methods) {
        return ExecuteAuto(Arrays.asList(methods));
    }

    /**
     * 自动开启线程去执行代码，并根据执行时间调整先后顺序
     *
     * @param methods
     */
    public static ThreadList<RunnableListAutoItem> ExecuteAuto(List<RunnableListAutoItem> methods) {
        // 获取历史执行时间
        int zeroCount = 0;
        long totalAvg = 0;
        long maxAvg = 0;

        // 没有执行过的列表
        List<RunnableListAutoItem> initList = new ArrayList<RunnableListAutoItem>();
        // 已经执行过的数据缓存，待排序
        Map<String, List<RunnableListAutoItem>> executeMap = new HashMap<String, List<RunnableListAutoItem>>();
        // 已执行过的列表
        List<RunnableListAutoItem> executedToList = new ArrayList<RunnableListAutoItem>();

        // 缓存的执行对象
        for (RunnableListAutoItem from : methods) {
            // 初始缓存执行对象
            if (!cache.containsKey(from.getKey())) {
                synchronized (cache) {
                    if (!cache.containsKey(from.getKey())) {
                        cache.put(from.getKey(), from.newBase());
                    }
                }
            }

            // 已经执行过的执行对象
            RunnableListAutoItem to = cache.get(from.getKey());
            if (to.getCount() == 0) {
                initList.add(from);
                zeroCount++;
            } else {
                if (!executeMap.containsKey(from.getKey())) {
                    executedToList.add(to);
                    executeMap.put(from.getKey(), new ArrayList<RunnableListAutoItem>());
                }
                executeMap.get(from.getKey()).add(from);
                totalAvg += to.getAvgTime();
                maxAvg = Math.max(to.getAvgTime(), maxAvg);
            }
        }

        // 将sort之后的结果，和当前执行的线程写入到 initList
        Collections.sort(executedToList);
        Collections.reverse(executedToList);
        for (RunnableListAutoItem to : executedToList) {
            List<RunnableListAutoItem> fromList = executeMap.get(to.getKey());
            initList.addAll(fromList);
        }

        // 计算线程数量
        maxAvg = Math.max(maxAvg, 1);
        int threadCount = zeroCount + (int) (totalAvg / maxAvg);
        if (threadCount > 2) {
            threadCount = threadCount - 1;
        }
        threadCount = Math.min(threadCount, methods.size());
        int finalThreadCount = threadCount;

        // 打印线程顺序
        if (IsLog) {
            for (RunnableListAutoItem item : initList) {
                if (item == null) {
                    continue;
                }
                RunnableListAutoItem toItem = cache.get(item.getKey());
                WriteLog("上次执行时间", toItem, finalThreadCount);
            }
        }

        // 开始线程列表
        ThreadList<RunnableListAutoItem> threadList = new ThreadList<RunnableListAutoItem>(finalThreadCount) {
            @Override
            public void run(RunnableListAutoItem item) {
                try {
                    item.execute();
                    if (IsLog) {
                        WriteLog("执行时间", item, finalThreadCount);
                    }
                } finally {
                    RunnableListAutoItem to = cache.get(item.getKey());
                    to.add(item);
                }
            }
        };

        // 添加到执行列表
        threadList.add(initList);
        // 等待线程执行完成
        threadList.waitRun();

        return threadList;
    }

    private static void WriteLog(String tag, RunnableListAutoItem item, int threadCount) {
        if (!IsLog) {
            return;
        }
        String msg = String.format("%s 类名: %s,%s,%s 次: %d  平均时间: %d 线程编号: %d 线程数量:%d", tag,
                item.getKey(), item.getClassType().getName(), item.getName(),
                item.getCount(), item.getAvgTime(),
                Thread.currentThread().getId(), threadCount);
        Log.info(RunnableListAuto.class, msg);
    }
}
